//+------------------------------------------------------------------+
//|                                                       Fincot.mqh |
//|                               Copyright 2014-2021, EarnForex.com |
//|                                              Developed by denkir |
//|                           https://login.mql5.com/ru/users/denkir |
//+------------------------------------------------------------------+
#property copyright "EarnForex.com"
#property link      "https://www.earnforex.com/guides/backtesting-strategies-based-on-commitments-of-traders/"
#property version   "1.01"

#define TFF_STORE_FOLDER     "COT\\fincot" // начало имени папки хранения файлов вида deacotxxxx
#define COT_MAIN_URL         "https://www.cftc.gov/MarketReports/CommitmentsofTraders/HistoricalCompressed/index.htm"

input bool InpReverse = false; // Reverse flag
input int StartYear = 2014;
input int EndYear = 2021;

//+------------------------------------------------------------------+
//|                 Перечисление типов трейдеров                     |
//+------------------------------------------------------------------+
enum ENUM_TFF_TRADER
{
    TFF_TRADER_DEALER,      // Дилеры
    TFF_TRADER_ASSET_MGR,   // Управляющие
    TFF_TRADER_LEV_MONEY,   // Хеджеры
    TFF_TRADER_OTHER_REPT,  // Прочие отчётные
    TFF_TRADER_NONREPT,     // Неотчётные
    TFF_TRADER_ALL,         // Все
};
//+------------------------------------------------------------------+
//|            Перечисление типов отражаемых показателей             |
//+------------------------------------------------------------------+
enum ENUM_TFF_TYPE_DATA
{
    //--- data
    TFF_TYPE_DATA_LONG,            // Лонги
    TFF_TYPE_DATA_SHORT,           // Шорты
    TFF_TYPE_DATA_SPREAD,          // Спред
    TFF_TYPE_DATA_OI,              // Открытый интерес

    //--- data change
    TFF_TYPE_DATA_LONG_CHG,        // Лонги,изм.
    TFF_TYPE_DATA_SHORT_CHG,       // Шорты,изм.
    TFF_TYPE_DATA_SPREAD_CHG,      // Спред,изм.
    TFF_TYPE_DATA_OI_CHG,          // Открытый интерес,изм.

    //--- percent of OI represented by
    TFF_TYPE_DATA_LONG_P_OI,       // Лонги,OI%
    TFF_TYPE_DATA_SHORT_P_OI,      // Шорты,OI%
    TFF_TYPE_DATA_SPREAD_P_OI,     // Спред,OI%

    //--- percent of  Total Long/Short/Spreading represented by
    TFF_TYPE_DATA_LONG_P_TOT,       // Лонги,T%
    TFF_TYPE_DATA_SHORT_P_TOT,      // Шорты,T%
    TFF_TYPE_DATA_SPREAD_P_TOT,     // Спред,T%
};
//+------------------------------------------------------------------+
//|            Перечисление типа входа/выхода                        |
//+------------------------------------------------------------------+
enum ENUM_IN_OUT
{
    IN_OUT_ENTER,     // вход
    IN_OUT_EXIT,      // выход
    IN_OUT_EXTRA_EXIT,// экстренный выход
};
//+------------------------------------------------------------------+
//|               Класс строки данных из отчета TFF                  |
//+------------------------------------------------------------------+
class CTFFRecord
{
    //+------------------------------------------------------------------+
    //=======================Члены-данные===============================|
    //+------------------------------------------------------------------+
public:
    //--- General
    bool              m_is_empty;          // flag
    string            m_tff_symbol;        // сurrency pair
    string            m_tff_ID;            // идентификатор инструмента
    datetime          m_tff_date;          // report date

    //---========================== Data
    double            m_tff_total_OI;      // total open interest
    //--- [0] - long,[1] - short,[2] - spreading
    double            m_tff_dealer[3];     // dealer position
    double            m_tff_asset_mgr[3];  // asset manager/institutional
    double            m_tff_lev_money[3];  // leveraged funds
    double            m_tff_other_rept[3]; // other reportables
    double            m_tff_nonrept[2];    // nonreportable positions
    //--- change
    double            m_tff_total_change_OI;      // total change in open interest
    double            m_tff_change_dealer[3];     // change in dealer position
    double            m_tff_change_asset_mgr[3];  // change in asset manager/institutional
    double            m_tff_change_lev_money[3];  // change in leveraged funds
    double            m_tff_change_other_rept[3]; // change in other reportables
    double            m_tff_change_nonrept[2];    // change in nonreportable positions
    //--- percent of OI represented by
    double            m_tff_OI_pct_dealer[3];     // dealer position
    double            m_tff_OI_pct_asset_mgr[3];  // asset manager/institutional
    double            m_tff_OI_pct_lev_money[3];  // leveraged funds
    double            m_tff_OI_pct_other_rept[3]; // other reportables
    double            m_tff_OI_pct_nonrept[2];    // nonreportable positions
    //--- total long/short contracts
    double            m_tff_total[3];             // total long/short/spreading contracts
    //--- percent of Total Long/Short/Spreading represented by
    double            m_tff_t_pct_dealer[3];     // dealer position
    double            m_tff_t_pct_asset_mgr[3];  // asset manager/institutional
    double            m_tff_t_pct_lev_money[3];  // leveraged funds
    double            m_tff_t_pct_other_rept[3]; // other reportables
    double            m_tff_t_pct_nonrept[2];    // nonreportable positions
    //---
    //+------------------------------------------------------------------+
    //=======================Члены-функции===============================|
    //+------------------------------------------------------------------+
public:
    void              CTFFRecord(void);   // конструктор
    void             ~CTFFRecord(void) {}; // деструктор
    //---
    void              CopyByValue(const CTFFRecord &_tff_record);
    void              ReverseLongShort(void);
};
//+------------------------------------------------------------------+
//|                              Конструктор                         |
//+------------------------------------------------------------------+
void CTFFRecord::CTFFRecord(void)
{
    this.m_is_empty = true;
    this.m_tff_symbol = NULL;
    this.m_tff_ID = NULL;
    this.m_tff_date = WRONG_VALUE;
    this.m_tff_total_OI = WRONG_VALUE;
    ArrayInitialize(this.m_tff_dealer, WRONG_VALUE);
    ArrayInitialize(this.m_tff_asset_mgr, WRONG_VALUE);
    ArrayInitialize(this.m_tff_lev_money, WRONG_VALUE);
    ArrayInitialize(this.m_tff_other_rept, WRONG_VALUE);
    ArrayInitialize(this.m_tff_nonrept, WRONG_VALUE);
//--- change
    this.m_tff_total_change_OI = WRONG_VALUE;
    ArrayInitialize(this.m_tff_change_dealer, WRONG_VALUE);
    ArrayInitialize(this.m_tff_change_asset_mgr, WRONG_VALUE);
    ArrayInitialize(this.m_tff_change_lev_money, WRONG_VALUE);
    ArrayInitialize(this.m_tff_change_other_rept, WRONG_VALUE);
    ArrayInitialize(this.m_tff_change_nonrept, WRONG_VALUE);
//--- percent of OI represented by
    ArrayInitialize(this.m_tff_OI_pct_dealer, WRONG_VALUE);
    ArrayInitialize(this.m_tff_OI_pct_asset_mgr, WRONG_VALUE);
    ArrayInitialize(this.m_tff_OI_pct_lev_money, WRONG_VALUE);
    ArrayInitialize(this.m_tff_OI_pct_other_rept, WRONG_VALUE);
    ArrayInitialize(this.m_tff_OI_pct_nonrept, WRONG_VALUE);
//--- total long/short contracts
    ArrayInitialize(this.m_tff_total, WRONG_VALUE);
//--- percent of  Total Long/Short/Spreading represented by
    ArrayInitialize(this.m_tff_t_pct_dealer, WRONG_VALUE);
    ArrayInitialize(this.m_tff_t_pct_asset_mgr, WRONG_VALUE);
    ArrayInitialize(this.m_tff_t_pct_lev_money, WRONG_VALUE);
    ArrayInitialize(this.m_tff_t_pct_other_rept, WRONG_VALUE);
    ArrayInitialize(this.m_tff_t_pct_nonrept, WRONG_VALUE);
}
//+------------------------------------------------------------------+
//|                  Копирование по значению                         |
//+------------------------------------------------------------------+
void CTFFRecord::CopyByValue(const CTFFRecord &_record)
{
    this.m_is_empty = _record.m_is_empty;
    this.m_tff_symbol = _record.m_tff_symbol;
    this.m_tff_ID = _record.m_tff_ID;
    this.m_tff_date = _record.m_tff_date;
    this.m_tff_total_OI = _record.m_tff_total_OI;
//---
    ArrayCopy(this.m_tff_dealer, _record.m_tff_dealer);
    ArrayCopy(this.m_tff_dealer, _record.m_tff_dealer);
    ArrayCopy(this.m_tff_asset_mgr, _record.m_tff_asset_mgr);
    ArrayCopy(this.m_tff_lev_money, _record.m_tff_lev_money);
    ArrayCopy(this.m_tff_other_rept, _record.m_tff_other_rept);
    ArrayCopy(this.m_tff_nonrept, _record.m_tff_nonrept);
//--- change
    this.m_tff_total_change_OI = _record.m_tff_total_change_OI;
    ArrayCopy(this.m_tff_change_dealer, _record.m_tff_change_dealer);
    ArrayCopy(this.m_tff_change_asset_mgr, _record.m_tff_change_asset_mgr);
    ArrayCopy(this.m_tff_change_lev_money, _record.m_tff_change_lev_money);
    ArrayCopy(this.m_tff_change_other_rept, _record.m_tff_change_other_rept);
    ArrayCopy(this.m_tff_change_nonrept, _record.m_tff_change_nonrept);
//--- percent of OI represented by
    ArrayCopy(this.m_tff_OI_pct_dealer, _record.m_tff_OI_pct_dealer);
    ArrayCopy(this.m_tff_OI_pct_asset_mgr, _record.m_tff_OI_pct_asset_mgr);
    ArrayCopy(this.m_tff_OI_pct_lev_money, _record.m_tff_OI_pct_lev_money);
    ArrayCopy(this.m_tff_OI_pct_other_rept, _record.m_tff_OI_pct_other_rept);
    ArrayCopy(this.m_tff_OI_pct_nonrept, _record.m_tff_OI_pct_nonrept);
//--- total long/short contracts
    ArrayCopy(this.m_tff_total, _record.m_tff_total);
//--- percent of  Total Long/Short/Spreading represented by
    ArrayCopy(this.m_tff_t_pct_dealer, _record.m_tff_t_pct_dealer);
    ArrayCopy(this.m_tff_t_pct_asset_mgr, _record.m_tff_t_pct_asset_mgr);
    ArrayCopy(this.m_tff_t_pct_lev_money, _record.m_tff_t_pct_lev_money);
    ArrayCopy(this.m_tff_t_pct_other_rept, _record.m_tff_t_pct_other_rept);
    ArrayCopy(this.m_tff_t_pct_nonrept, _record.m_tff_t_pct_nonrept);
};
//+------------------------------------------------------------------+
//|                Reverse Long & Short data                         |
//+------------------------------------------------------------------+
void CTFFRecord::ReverseLongShort(void)
{
    double temp_arr[];

//--- [0] - long,[1] - short,[2] - spreading
    ArrayCopy(temp_arr, this.m_tff_dealer);
    this.m_tff_dealer[0] = temp_arr[1];
    this.m_tff_dealer[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_asset_mgr);
    this.m_tff_asset_mgr[0] = temp_arr[1];
    this.m_tff_asset_mgr[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_lev_money);
    this.m_tff_lev_money[0] = temp_arr[1];
    this.m_tff_lev_money[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_other_rept);
    this.m_tff_other_rept[0] = temp_arr[1];
    this.m_tff_other_rept[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_nonrept);
    this.m_tff_nonrept[0] = temp_arr[1];
    this.m_tff_nonrept[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_change_dealer);
    this.m_tff_change_dealer[0] = temp_arr[1];
    this.m_tff_change_dealer[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_change_asset_mgr);
    this.m_tff_change_asset_mgr[0] = temp_arr[1];
    this.m_tff_change_asset_mgr[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_change_lev_money);
    this.m_tff_change_lev_money[0] = temp_arr[1];
    this.m_tff_change_lev_money[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_change_other_rept);
    this.m_tff_change_other_rept[0] = temp_arr[1];
    this.m_tff_change_other_rept[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_change_nonrept);
    this.m_tff_change_nonrept[0] = temp_arr[1];
    this.m_tff_change_nonrept[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_OI_pct_dealer);
    this.m_tff_OI_pct_dealer[0] = temp_arr[1];
    this.m_tff_OI_pct_dealer[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_OI_pct_asset_mgr);
    this.m_tff_OI_pct_asset_mgr[0] = temp_arr[1];
    this.m_tff_OI_pct_asset_mgr[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_OI_pct_lev_money);
    this.m_tff_OI_pct_lev_money[0] = temp_arr[1];
    this.m_tff_OI_pct_lev_money[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_OI_pct_other_rept);
    this.m_tff_OI_pct_other_rept[0] = temp_arr[1];
    this.m_tff_OI_pct_other_rept[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_OI_pct_nonrept);
    this.m_tff_OI_pct_nonrept[0] = temp_arr[1];
    this.m_tff_OI_pct_nonrept[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_total);
    this.m_tff_total[0] = temp_arr[1];
    this.m_tff_total[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_t_pct_dealer);
    this.m_tff_t_pct_dealer[0] = temp_arr[1];
    this.m_tff_t_pct_dealer[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_t_pct_asset_mgr);
    this.m_tff_t_pct_asset_mgr[0] = temp_arr[1];
    this.m_tff_t_pct_asset_mgr[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_t_pct_lev_money);
    this.m_tff_t_pct_lev_money[0] = temp_arr[1];
    this.m_tff_t_pct_lev_money[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_t_pct_other_rept);
    this.m_tff_t_pct_other_rept[0] = temp_arr[1];
    this.m_tff_t_pct_other_rept[1] = temp_arr[0];
//---
    ArrayCopy(temp_arr, this.m_tff_t_pct_nonrept);
    this.m_tff_t_pct_nonrept[0] = temp_arr[1];
    this.m_tff_t_pct_nonrept[1] = temp_arr[0];
//---
    ArrayFree(temp_arr);
}
//+------------------------------------------------------------------+
//|                 Signal Data structure                            |
//+------------------------------------------------------------------+
struct SSignalData
{
    //--- members
    int               idx;        // index
    datetime          start_date; // start date
    datetime          curr_date;  // current date
    datetime          tff_date;   // report date
    uint              id;         // ID
    ENUM_ORDER_TYPE   ord_type;   // order type
    double            OI_val;     // Open Interest value
    double            dealer_vals[2]; // dealer_vals[0]-long,dealer_vals[1]-short
    double            asset_vals[2];
    double            lev_vals[2];
    double            nonrept_vals[2];
    //---
    CTFFRecord        curr_record; // current record
    CTFFRecord        prev_record; // previous record

    //--- methods
    void              Init(void);
    void              CopyByValue(const SSignalData &_source_sig);
};
//+------------------------------------------------------------------+
//|                    Initialization                                |
//+------------------------------------------------------------------+
void SSignalData::Init(void)
{
    this.idx = WRONG_VALUE;
    this.id = WRONG_VALUE;
    this.ord_type = WRONG_VALUE;
    this.start_date = this.curr_date = this.tff_date = WRONG_VALUE;
    this.OI_val = 0.0;
    ArrayInitialize(dealer_vals, 0.0);
    ArrayInitialize(asset_vals, 0.0);
    ArrayInitialize(lev_vals, 0.0);
    ArrayInitialize(nonrept_vals, 0.0);
}
//+------------------------------------------------------------------+
//|                    Copy by value                                 |
//+------------------------------------------------------------------+
void SSignalData::CopyByValue(const SSignalData &_source_sig)
{
    this.idx = _source_sig.idx;
    this.start_date = _source_sig.start_date;
    this.curr_date = _source_sig.curr_date;
    this.tff_date = _source_sig.tff_date;
//---
    this.id = _source_sig.id;
    this.ord_type = _source_sig.ord_type;
    this.OI_val = _source_sig.OI_val;
    ArrayCopy(this.dealer_vals, _source_sig.dealer_vals); // dealer_vals[0]-long,dealer_vals[1]-short
    ArrayCopy(this.asset_vals, _source_sig.asset_vals);
    ArrayCopy(this.lev_vals, _source_sig.lev_vals);
    ArrayCopy(this.nonrept_vals, _source_sig.nonrept_vals);
//---
    this.curr_record.CopyByValue(_source_sig.curr_record);
    this.prev_record.CopyByValue(_source_sig.prev_record);
}
//+------------------------------------------------------------------+
//|                    Класс отчёта TFF                              |
//+------------------------------------------------------------------+
class CTFFReport
{
    //+------------------------------------------------------------------+
    //=======================Члены-данные===============================|
    //+------------------------------------------------------------------+
private:
    CTFFRecord        m_records_arr[];  // массив записей
    uint              m_records_size;   // число записей TFF-отчёта
    string            m_tff_id_symbol;  // идентификатор инструмента в TFF-отчёте, поле "CFTC Contract Market Code"
    string            m_mt_symbol;      // символ инструмента в терминале
    //---
    SSignalData       m_signal;         // signal data

    //+------------------------------------------------------------------+
    //=======================Члены-функции===============================|
    //+------------------------------------------------------------------+
public:
    //--- конструктор и деструктор
    void              CTFFReport(void);     // конструктор
    void             ~CTFFReport(void);     // деструктор
    //--- сервис
    bool              Init(string _term_symbol = NULL); // инициализация
    bool              LoadFile(string _filename);     // загрузка данных из внешнего файла
    bool              DefineIdSymbol();               //определение идинтификатора в отчете сот
    bool              GetString(int handle, CTFFRecord &_arr); //получить запись csv из файла
    string            GetFieldCommaDelimited(string _source_string, const uint _field_num); //получить поле из csv строки
    //--- получить данные из отчета
    double            At(datetime _date, ENUM_TFF_TRADER _type_trader, ENUM_TFF_TYPE_DATA _typedata);
    //--- сигналы
    ENUM_ORDER_TYPE   CheckOpenSignal(const datetime _date, const uint _signal_id);
    ENUM_ORDER_TYPE   CheckCloseSignal(const datetime _date);
    void              SetSignal(const SSignalData &_signal);
    void              GetSignal(SSignalData &_out_signal);
    void              ResetSignal(void);
    uint              GetRecordsSize(void)
    {
        return this.m_records_size;
    };
    void              GetRecords(CTFFRecord &_records_arr[]);
    //--- приватные методы
private:
    ENUM_ORDER_TYPE   DefineSignal(const double &_vals[], const ENUM_IN_OUT bEntry);
    bool              CheckReverse(void);   // проверка в записи на изменение местами long и short
    void              ProcessReverse(void); // изменить местами long и short
    //---
    int               GetIdxByDate(datetime _date);
    bool              RefreshSignalData(const datetime _date, const uint _signal_id, bool _for_open = true);
    uint              AdjustSignalNum(uint _curr_sig_id);
};
//+------------------------------------------------------------------+
//|                       Конструктор                                |
//+------------------------------------------------------------------+
void CTFFReport::CTFFReport(void)
{
    this.m_records_size = 0; // число записей TFF-отчёта
    this.m_tff_id_symbol = NULL; // идентификатор инструмента в TFF-отчёте, поле "CFTC Contract Market Code"
    this.m_mt_symbol = NULL; // символ инструмента в терминале
}
//+------------------------------------------------------------------+
//|                        Деструктор                                |
//+------------------------------------------------------------------+
void CTFFReport::~CTFFReport(void)
{
    this.m_records_size = 0;
    this.m_tff_id_symbol = NULL;
    this.m_mt_symbol = NULL;
    ArrayFree(this.m_records_arr);
}
//+------------------------------------------------------------------+
//| Получить показатель из отчета COT                                |
//| Вход  : time - дата, на которую необходимо получить данные       |
//|         _type_trader - тип группы трейдеров                      |
//|         _typedata - тип возращаемых данных                       |
//| Выход : значение показателя                                      |
//+------------------------------------------------------------------+
double CTFFReport::At(datetime _date, ENUM_TFF_TRADER _type_trader, ENUM_TFF_TYPE_DATA _typedata)
{
    int data_idx = this.GetIdxByDate(_date); // this.m_records_size-1;
    if (data_idx == WRONG_VALUE)
        return WRONG_VALUE;
// после цикла найден элемент массива с необходимыми данными
    CTFFRecord curr_record;
    curr_record.CopyByValue(this.m_records_arr[data_idx]); // выбранная запись
    double ret_data = WRONG_VALUE;
//---
    if (_typedata == TFF_TYPE_DATA_OI && _type_trader == TFF_TRADER_ALL)
        ret_data = curr_record.m_tff_total_OI;
//---
    if (_typedata == TFF_TYPE_DATA_OI_CHG && _type_trader == TFF_TRADER_ALL)
        ret_data = curr_record.m_tff_total_change_OI;
//---
    if (_typedata >= TFF_TYPE_DATA_LONG && _typedata <= TFF_TYPE_DATA_SPREAD)
    {
        int type_idx = _typedata;
        switch(_type_trader)
        {
        case TFF_TRADER_DEALER:
            ret_data = curr_record.m_tff_dealer[type_idx];
            break;
        case TFF_TRADER_ASSET_MGR:
            ret_data = curr_record.m_tff_asset_mgr[type_idx];
            break;
        case TFF_TRADER_LEV_MONEY:
            ret_data = curr_record.m_tff_lev_money[type_idx];
            break;
        case TFF_TRADER_OTHER_REPT:
            ret_data = curr_record.m_tff_other_rept[type_idx];
            break;
        case TFF_TRADER_NONREPT:
            if (type_idx == 2)
            {
                Print("No spread data for nonreportable positions!");
                ret_data = WRONG_VALUE;
            }
            else
                ret_data = curr_record.m_tff_nonrept[type_idx];
            break;
        case TFF_TRADER_ALL:
            ret_data = curr_record.m_tff_total[type_idx];
            break;
        }
    }
//---
    else if (_typedata >= TFF_TYPE_DATA_LONG_CHG && _typedata <= TFF_TYPE_DATA_SPREAD_CHG)
    {
        int type_idx = _typedata - 4; // todo
        switch(_type_trader)
        {
        case TFF_TRADER_DEALER:
            ret_data = curr_record.m_tff_change_dealer[type_idx];
            break;
        case TFF_TRADER_ASSET_MGR:
            ret_data = curr_record.m_tff_change_asset_mgr[type_idx];
            break;
        case TFF_TRADER_LEV_MONEY:
            ret_data = curr_record.m_tff_change_lev_money[type_idx];
            break;
        case TFF_TRADER_OTHER_REPT:
            ret_data = curr_record.m_tff_change_other_rept[type_idx];
            break;
        case TFF_TRADER_NONREPT:
            if (type_idx == 2)
            {
                Print("No spread change data for nonreportable positions!");
                ret_data = WRONG_VALUE;
            }
            else
                ret_data = curr_record.m_tff_change_nonrept[type_idx];
            break;
        }
    }
//---
    else if (_typedata >= TFF_TYPE_DATA_LONG_P_OI && _typedata <= TFF_TYPE_DATA_SPREAD_P_OI)
    {
        int type_idx = _typedata - 8; // todo
        switch(_type_trader)
        {
        case TFF_TRADER_DEALER:
            ret_data = curr_record.m_tff_OI_pct_dealer[type_idx];
            break;
        case TFF_TRADER_ASSET_MGR:
            ret_data = curr_record.m_tff_OI_pct_asset_mgr[type_idx];
            break;
        case TFF_TRADER_LEV_MONEY:
            ret_data = curr_record.m_tff_OI_pct_lev_money[type_idx];
            break;
        case TFF_TRADER_OTHER_REPT:
            ret_data = curr_record.m_tff_OI_pct_other_rept[type_idx];
            break;
        case TFF_TRADER_NONREPT:
            if (type_idx == 2)
            {
                Print("No  percent of OI spread data for nonreportable positions!");
                ret_data = WRONG_VALUE;
            }
            else
                ret_data = curr_record.m_tff_OI_pct_nonrept[type_idx];
            break;
        }
    }
//---
    else if (_typedata >= TFF_TYPE_DATA_LONG_P_TOT && _typedata <= TFF_TYPE_DATA_SPREAD_P_TOT)
    {
        int type_idx = _typedata - 11; // todo
        switch(_type_trader)
        {
        case TFF_TRADER_DEALER:
            ret_data = curr_record.m_tff_t_pct_dealer[type_idx];
            break;
        case TFF_TRADER_ASSET_MGR:
            ret_data = curr_record.m_tff_t_pct_asset_mgr[type_idx];
            break;
        case TFF_TRADER_LEV_MONEY:
            ret_data = curr_record.m_tff_t_pct_lev_money[type_idx];
            break;
        case TFF_TRADER_OTHER_REPT:
            ret_data = curr_record.m_tff_t_pct_other_rept[type_idx];
            break;
        case TFF_TRADER_NONREPT:
            if (type_idx == 2)
            {
                Print("No percent of Total spread data for nonreportable positions!");
                ret_data = WRONG_VALUE;
            }
            else
                ret_data = curr_record.m_tff_t_pct_nonrept[type_idx];
            break;
        }
    }
//---
    return ret_data;
}
//+------------------------------------------------------------------+
//|                     Сигнал к открытию                            |
//+------------------------------------------------------------------+
ENUM_ORDER_TYPE CTFFReport::CheckOpenSignal(const datetime _date, const uint _signal_id)
{
    //
    if (_signal_id < 1 || _signal_id > 32)
    {
        Print("Wrong signal id.");
        return WRONG_VALUE;
    }
//--- check the active signal
    if (this.m_signal.ord_type != WRONG_VALUE)
    {
        Print("Сигнал открытия уже задан: " + EnumToString(this.m_signal.ord_type));
        return this.m_signal.ord_type;
    }
//---
    if (!this.RefreshSignalData(_date, _signal_id))
        return WRONG_VALUE;
//---
    ENUM_ORDER_TYPE signal_ord_type = WRONG_VALUE;

//--- check signals
    ENUM_ORDER_TYPE dealer_asset_signal = WRONG_VALUE;
    //ENUM_ORDER_TYPE dealer_asset_lev_signal=WRONG_VALUE;

    ENUM_ORDER_TYPE dealer_signal = WRONG_VALUE;
    ENUM_ORDER_TYPE asset_signal = WRONG_VALUE;
    ENUM_ORDER_TYPE nonrept_signal = WRONG_VALUE;

//--- category trade signal [start]
    /*
    const double _long_val=_vals1[0]+_vals2[0];
       const double _short_val=_vals1[1]+_vals2[1];
    */
    double Vals[2];
    double long_val = this.m_signal.dealer_vals[0] + this.m_signal.asset_vals[0];
    double short_val = this.m_signal.dealer_vals[1] + this.m_signal.asset_vals[1];
    Vals[0] = long_val;
    Vals[1] = short_val;
    dealer_asset_signal = this.DefineSignal(Vals, IN_OUT_ENTER);
    //long_val+=this.m_signal.lev_vals[0];
    //short_val+=this.m_signal.lev_vals[1];
    //Vals[0]=long_val;
    //Vals[1]=short_val;
    //dealer_asset_lev_signal=this.DefineSignal(Vals,IN_OUT_ENTER);
//---
    dealer_signal = this.DefineSignal(this.m_signal.dealer_vals, IN_OUT_ENTER);
    asset_signal = this.DefineSignal(this.m_signal.asset_vals, IN_OUT_ENTER);
    nonrept_signal = this.DefineSignal(this.m_signal.nonrept_vals, IN_OUT_ENTER);

//--- category trade signal [end]

//--- Aggregate trade signal
    uint curr_signal_id = this.AdjustSignalNum(_signal_id);

//--- определить open сигнал по id
    switch(curr_signal_id) // notation for buy signal
    {
    case 1: // Dealer longs — positive change AND Dealer shorts — negative change
        signal_ord_type = dealer_signal;
        break;
    case 2: // (Dealer+AssetMan/Inst) longs — positive change AND (Dealer+AssetMan/Inst) shorts — negative change
        signal_ord_type = dealer_asset_signal;
        break;
    case 3: // Dealer longs % of total longs — positive change AND Dealer shorts % of total shorts — negative change
        signal_ord_type = dealer_signal;
        break;
    case 4: // (Dealer longs AND AssetMan/Inst longs — positive change) AND (Dealer shorts AND AssetMan/Inst shorts — negative change
        if (dealer_signal == asset_signal)
            signal_ord_type = dealer_signal;
        break;
    case 5: // (Dealer longs % of total longs AND AssetMan/Inst longs % of total longs — positive change) AND (Dealer shorts % of total shorts AND AssetMan/Inst shorts % of total shorts — negative change)
    {
        if (dealer_signal == asset_signal)
            signal_ord_type = dealer_signal;
        break;
    }
    //---
    case 11: // Dealer longs % of OI — positive change AND Dealer shorts % of OI — negative change
        signal_ord_type = dealer_signal;
        break;
    case 12: // (Dealer longs % of OI AND AssetMan/Inst longs % of OI — positive change) AND (Dealer shorts % of OI AND AssetMan/Inst shorts % of OI — negative change)
    {
        if (dealer_signal == asset_signal)
            signal_ord_type = dealer_signal;
        break;
    }
    case 13: // NonRep longs — positive change AND NonRep shorts — negative change
    {
        signal_ord_type = nonrept_signal;
        break;
    }
    case 14: // Dealer longs % of OI — positive change AND Dealer shorts % of OI — negative change
        signal_ord_type = dealer_signal;
        break;
    case 15: // (Dealer longs % of OI AND AssetMan/Inst longs % of OI — positive change) AND (Dealer shorts % of OI AND AssetMan/Inst shorts % of OI — negative change)
    {
        if (dealer_signal == asset_signal)
            signal_ord_type = dealer_signal;
        break;
    }
        /*      case 16: // (Dealer % of longs AND AssetMan/Inst % of longs AND LevFunds % of longs — positive change) AND (Dealer % of shorts AND AssetMan/Inst % of shorts AND LevFunds % of shorts — negative change)
                {
                 if (dealer_signal==asset_signal)
                    if (asset_signal==lev_signal)
                       signal_ord_type=dealer_signal;
                 break;
                }*/
    }
//---
    if (signal_ord_type != WRONG_VALUE)
        if (_signal_id > 16) // check whether total change in OI is positive
        {
            double curr_t_OI_chg = this.m_signal.OI_val;
            if (curr_t_OI_chg < 0.0)
                signal_ord_type = WRONG_VALUE; // reset signal
        }
//--- определить тип ордера
    if (signal_ord_type != WRONG_VALUE)
    {
        if (InpReverse)
        {
            signal_ord_type = (signal_ord_type == ORDER_TYPE_BUY) ? ORDER_TYPE_SELL : ORDER_TYPE_BUY;
            Print("Использован реверс сигнала.");
            Print("Новый сигнал: " + EnumToString(signal_ord_type));
        }
        this.m_signal.ord_type = signal_ord_type;
    }

//---
    return signal_ord_type;
}
//+------------------------------------------------------------------+
//|                    Сигнал к закрытию                             |
//+------------------------------------------------------------------+
ENUM_ORDER_TYPE CTFFReport::CheckCloseSignal(const datetime _date)
{
    if (this.m_signal.ord_type == WRONG_VALUE)
    {
        Print("Первичный сигнал не задан.");
        return WRONG_VALUE;
    }
//--- DELETE
//Print("Дата сигнала: "+TimeToString(this.m_signal.date));
    // Print("Текущая дата: "+TimeToString(_date));
//--- DELETE

    if (this.m_signal.start_date >= (_date - 5 * PeriodSeconds(PERIOD_D1)))
        return WRONG_VALUE;

    if (!this.RefreshSignalData(_date, this.m_signal.id, false))
        return WRONG_VALUE;

    ENUM_ORDER_TYPE signal_ord_type = WRONG_VALUE;

//--- check signals
    ENUM_ORDER_TYPE dealer_asset_signal = WRONG_VALUE;
    //ENUM_ORDER_TYPE dealer_asset_lev_signal=WRONG_VALUE;
    ENUM_ORDER_TYPE dealer_signal = WRONG_VALUE;
    ENUM_ORDER_TYPE asset_signal = WRONG_VALUE;
    //ENUM_ORDER_TYPE lev_signal=WRONG_VALUE;
    ENUM_ORDER_TYPE nonrept_signal = WRONG_VALUE;
//---
    ENUM_ORDER_TYPE dealer_asset_ex_signal = WRONG_VALUE;
    //ENUM_ORDER_TYPE dealer_asset_lev_ex_signal=WRONG_VALUE;
    ENUM_ORDER_TYPE dealer_ex_signal = WRONG_VALUE;
    ENUM_ORDER_TYPE asset_ex_signal = WRONG_VALUE;
    //ENUM_ORDER_TYPE lev_ex_signal=WRONG_VALUE;
    ENUM_ORDER_TYPE nonrept_ex_signal = WRONG_VALUE;

//--- category trade signal (closing)
    /*
    enum ENUM_IN_OUT
      {
       IN_OUT_ENTER,// вход
       IN_OUT_EXIT, // выход
       IN_OUT_EXTRA_EXIT,// экстренный выход
      };
    */

//--- ordinary exit
    ENUM_IN_OUT b_entry1 = IN_OUT_EXIT;
    ENUM_IN_OUT b_entry2 = IN_OUT_EXTRA_EXIT;
    double Vals[2];
    double long_val = this.m_signal.dealer_vals[0] + this.m_signal.asset_vals[0];
    double short_val = this.m_signal.dealer_vals[1] + this.m_signal.asset_vals[1];
    Vals[0] = long_val;
    Vals[1] = short_val;
    dealer_asset_signal = this.DefineSignal(Vals, b_entry1);
    dealer_asset_ex_signal = this.DefineSignal(Vals, b_entry2);
//---
    /*   long_val+=this.m_signal.lev_vals[0];
       short_val+=this.m_signal.lev_vals[1];
       Vals[0]=long_val;
       Vals[1]=short_val;
       dealer_asset_lev_signal=this.DefineSignal(Vals,b_entry1);
       dealer_asset_lev_ex_signal=this.DefineSignal(Vals,b_entry2);*/
//---

    dealer_signal = this.DefineSignal(this.m_signal.dealer_vals, b_entry1);
    asset_signal = this.DefineSignal(this.m_signal.asset_vals, b_entry1);
    nonrept_signal = this.DefineSignal(this.m_signal.nonrept_vals, b_entry1);

//--- extraordinary exit
    dealer_ex_signal = this.DefineSignal(this.m_signal.dealer_vals, b_entry2);
    asset_ex_signal = this.DefineSignal(this.m_signal.asset_vals, b_entry2);
    nonrept_ex_signal = this.DefineSignal(this.m_signal.nonrept_vals, b_entry2);

//--- Aggregate trade signal
    uint _signal_id = this.m_signal.id;

//--- определить close сигнал по id
    switch(_signal_id)
    {
    case 1: // extra [Dealer longs — negative change OR Dealer shorts — positive change]
        // ord: reverse enter
    {
        signal_ord_type = dealer_ex_signal;
        break;
    }
    case 2: // extra [(Dealer+AssetMan/Inst) longs — negative change OR (Dealer+AssetMan/Inst) shorts — positive change]
        // ord: reverse enter
    {
        signal_ord_type = dealer_asset_ex_signal;
        break;
    }
    case 3: // extra [Dealer longs % of total longs — negative change OR Dealer shorts % of total shorts — positive change]
        // ord: reverse enter
    {
        signal_ord_type = dealer_ex_signal;
        break;
    }
    case 4: // extra [(Dealer longs OR AssetMan/Inst longs — negative change) OR (Dealer shorts OR AssetMan/Inst shorts — positive change)]
        // ord: reverse enter
    {
        if (dealer_ex_signal == this.m_signal.ord_type || asset_ex_signal == this.m_signal.ord_type)
            signal_ord_type = this.m_signal.ord_type;
        break;
    }
    case 5: // extra [(Dealer longs % of total longs OR AssetMan/Inst longs % of total longs — negative change) OR (Dealer shorts % of total shorts OR AssetMan/Inst shorts % of total shorts — positive change)]
        // ord: reverse enter
    {
        if (dealer_ex_signal == this.m_signal.ord_type || asset_ex_signal == this.m_signal.ord_type)
            signal_ord_type = this.m_signal.ord_type;
        break;
    }
    case 6: // ord: reverse enter
    {
        signal_ord_type = dealer_signal;
        break;
    }
    case 7: // ord: reverse enter
    {
        signal_ord_type = dealer_asset_signal;
        break;
    }
    case 8: // ord: reverse enter
    {
        signal_ord_type = dealer_signal;
        break;
    }
    case 9: // ord: reverse enter
    {
        if (dealer_signal == asset_signal)
            signal_ord_type = dealer_signal;
        break;
    }
    case 10: // ord: reverse enter
    {
        if (dealer_signal == asset_signal)
            signal_ord_type = dealer_signal;
        break;
    }
    //---
    case 11: // extra [Dealer % of longs — negative change OR Dealer % of shorts — positive change]
        // ord: reverse enter
        signal_ord_type = dealer_ex_signal;
        break;
    case 12: // extra [(Dealer % of longs OR AssetMan/Inst % of longs — negative change) OR (Dealer % of shorts OR AssetMan/Inst % of shorts — positive change)]
        // ord: reverse enter
    {
        if (dealer_ex_signal == this.m_signal.ord_type || asset_ex_signal == this.m_signal.ord_type)
            signal_ord_type = this.m_signal.ord_type;
        break;
    }
    case 13: // extra [NonRep longs — negative change OR NonRep shorts — positive change]
        // ord: reverse enter
    {
        signal_ord_type = nonrept_ex_signal;
        break;
    }
    case 14:
        // ord: reverse enter
        signal_ord_type = dealer_signal;
        break;
    case 15:
        // ord: reverse enter
    {
        if (dealer_signal == asset_signal)
            signal_ord_type = dealer_signal;
        break;
    }
    case 16:
        // ord: reverse enter
    {
        signal_ord_type = nonrept_signal;
        break;
    }
    case 17: // extra [Dealer longs — negative change OR Dealer shorts — positive change]
        // ord: reverse enter
    {
        signal_ord_type = dealer_ex_signal;
        if (signal_ord_type != this.m_signal.ord_type)
            if (this.m_signal.OI_val > 0.0)
                signal_ord_type = dealer_signal;
        break;
    }
    case 18: // extra [(Dealer+AssetMan/Inst) longs — negative change OR (Dealer+AssetMan/Inst) shorts — positive change]
    {
        signal_ord_type = dealer_asset_ex_signal;
        if (signal_ord_type != this.m_signal.ord_type)
            if (this.m_signal.OI_val > 0.0)
                signal_ord_type = dealer_asset_signal;
        break;
    }
    case 19: // extra [(Dealer+AssetMan/Inst + LevFunds) longs — negative change OR (Dealer+AssetMan/Inst+LevFunds) shorts — positive change]
    {
        signal_ord_type = dealer_ex_signal;
        if (signal_ord_type != this.m_signal.ord_type)
            if (this.m_signal.OI_val > 0.0)
                signal_ord_type = dealer_signal;
        break;
    }
    case 20: // extra [(Dealer longs OR AssetMan/Inst longs — negative change) OR (Dealer shorts OR AssetMan/Inst shorts — positive change]
    {
        if (dealer_ex_signal == this.m_signal.ord_type || asset_ex_signal == this.m_signal.ord_type)
            signal_ord_type = this.m_signal.ord_type;
        break;
    }
    case 21: // extra [(Dealer longs OR AssetMan/Inst longs OR LevFunds longs — negative change) OR (Dealer shorts OR AssetMan/Inst shorts OR LevFunds shorts — positive change]
    {
        if (dealer_ex_signal == this.m_signal.ord_type || asset_ex_signal == this.m_signal.ord_type)
            signal_ord_type = this.m_signal.ord_type;
        break;
    }
    case 22:
    {
        if (this.m_signal.OI_val > 0.0)
            signal_ord_type = dealer_signal;
        break;
    }
    case 23:
    {
        if (this.m_signal.OI_val > 0.0)
            signal_ord_type = dealer_asset_signal;
        break;
    }
    case 24:
    {
        if (this.m_signal.OI_val > 0.0)
            signal_ord_type = dealer_signal;
        break;
    }
    case 25:
    {
        if (dealer_signal == asset_signal)
            signal_ord_type = dealer_signal;
        break;
    }
    case 26:
    {
        if (dealer_signal == asset_signal)
            signal_ord_type = dealer_signal;
        break;
    }
    case 27: // extra [Dealer % of longs — negative change OR Dealer % of shorts — positive change]
    {
        signal_ord_type = dealer_ex_signal;
        break;
    }
    case 28: // extra [(Dealer % of longs OR AssetMan/Inst % of longs — negative change) OR (Dealer % of shorts OR AssetMan/Inst % of shorts — positive change)]
    {
        if (dealer_ex_signal == this.m_signal.ord_type || asset_ex_signal == this.m_signal.ord_type)
            signal_ord_type = this.m_signal.ord_type;
        break;
    }
    case 29: // extra [(Dealer % of longs OR AssetMan/Inst % of longs OR LevFunds % of longs — negative change) OR (Dealer % of shorts OR AssetMan/Inst % of shorts OR LevFunds % of shorts — positive chang]
    {
        signal_ord_type = nonrept_ex_signal;
        break;
    }
    case 30:
    {
        if (this.m_signal.OI_val > 0.0)
            signal_ord_type = dealer_signal;
        break;
    }
    case 31:
    {
        if (this.m_signal.OI_val > 0.0)
            if (dealer_signal == asset_signal)
                signal_ord_type = dealer_signal;
        break;
    }
    case 32:
    {
        if (this.m_signal.OI_val > 0.0)
            signal_ord_type = nonrept_signal;
        break;
    }
    }
//---
    /* TODO
       if (signal_ord_type!=WRONG_VALUE)
          if (_signal_id>16) // if total change in OI is positive
            {
             double curr_t_OI_chg=this.m_signal.OI_val;
             if (curr_t_OI_chg<0.0)
                signal_ord_type=WRONG_VALUE; // reset signal
            }
            */
//--- if the signals differ - reset
    if (signal_ord_type != WRONG_VALUE)
        if (signal_ord_type != this.m_signal.ord_type)
            signal_ord_type = WRONG_VALUE;
//---
    return signal_ord_type;
}
//+------------------------------------------------------------------+
//|                    Задать свойства сигнала                       |
//+------------------------------------------------------------------+
void CTFFReport::SetSignal(const SSignalData &_signal)
{
    this.m_signal.CopyByValue(_signal);
}
//+------------------------------------------------------------------+
//|                    Получить свойства сигнала                     |
//+------------------------------------------------------------------+
void CTFFReport::GetSignal(SSignalData &_out_signal)
{
    _out_signal.CopyByValue(this.m_signal);
}
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CTFFReport::ResetSignal(void)
{
    this.m_signal.Init();
}
//+------------------------------------------------------------------+
//|                   Получить массив записей                        |
//+------------------------------------------------------------------+
void CTFFReport::GetRecords(CTFFRecord &_records_arr[])
{
    uint arr_new_size = this.m_records_size;
//---
    ArrayResize(_records_arr, this.m_records_size);
//---
    for(uint rec_idx = 0; rec_idx < arr_new_size; rec_idx++)
        _records_arr[rec_idx].CopyByValue(this.m_records_arr[rec_idx]);
}
//+------------------------------------------------------------------+
//|       Установить торговый сигнал для открытия/закрытия           |
//+------------------------------------------------------------------+
ENUM_ORDER_TYPE CTFFReport::DefineSignal(const double &_vals[], const ENUM_IN_OUT bEntry)
{
    /*
        bEntry = IN_OUT_ENTER - open;
        bEntry = IN_OUT_EXIT,IN_OUT_EXTRA_EXIT - close.
      */
    ENUM_ORDER_TYPE signal = WRONG_VALUE;
    const double _long_val = _vals[0];
    const double _short_val = _vals[1];

//---
    switch(bEntry)
    {
    case IN_OUT_ENTER:
    {
        if (_long_val > 0.0 && _short_val < 0.0)
            signal = ORDER_TYPE_BUY; // условие для покупок
        else if (_long_val < 0.0 && _short_val > 0.0)
            signal = ORDER_TYPE_SELL; // условие для продаж
        //---
        break;
    }
    //---
    case IN_OUT_EXIT:
    {
        if (_long_val > 0.0 && _short_val < 0.0)
            signal = ORDER_TYPE_SELL; // условие для покупок
        else if (_long_val < 0.0 && _short_val > 0.0)
            signal = ORDER_TYPE_BUY; // условие для продаж
        //---
        break;
    }
    //---
    case IN_OUT_EXTRA_EXIT:
    {
        ENUM_ORDER_TYPE curr_open_ord_type = this.m_signal.ord_type;
        if (curr_open_ord_type == ORDER_TYPE_SELL)
        {
            if (_long_val > 0.0 || _short_val < 0.0)
                signal = ORDER_TYPE_SELL;
        }
        else if (curr_open_ord_type == ORDER_TYPE_BUY)
        {
            if (_long_val < 0.0 || _short_val > 0.0)
                signal = ORDER_TYPE_BUY;
        }
        //---
        break;
    }
    }
//---
    return signal;
}
//+------------------------------------------------------------------+
//|            Установить торговый сигнал для закрытия               |
//+------------------------------------------------------------------+
/*ENUM_ORDER_TYPE CTFFReport::DefineCloseSignal(const double &_vals[],const bool bEntry)
  {
   ENUM_ORDER_TYPE signal=WRONG_VALUE;
   const double _long_val=_vals[0];
   const double _short_val=_vals[1];
//--- signal
   if (_long_val>0.0 || _short_val<0.0)
      signal=(bEntry)?ORDER_TYPE_BUY:ORDER_TYPE_SELL; // условие для покупок
   else if (_long_val<0.0 || _short_val>0.0)
      signal=(bEntry)?ORDER_TYPE_SELL:ORDER_TYPE_BUY; // условие для продаж
//---
   return signal;
  }*/
//+------------------------------------------------------------------+
//|       Проверка в записи на изменение местами long и short        |
//+------------------------------------------------------------------+
bool CTFFReport::CheckReverse(void)
{
    bool IsReverse = false;
//---
    string base_exchg = StringSubstr(this.m_mt_symbol, 0, 3);
    if (base_exchg == "USD")
    {
        IsReverse = true;
        Print("Reverse long and short for the current report symbol: " + this.m_mt_symbol);
    }
    return IsReverse;
}
//+------------------------------------------------------------------+
//|         Изменить местами long и short в массивах данных          |
//+------------------------------------------------------------------+
void CTFFReport::ProcessReverse(void)
{
    /*
        m_records_arr[];  // массив записей
       uint              m_records_size;       // число записей TFF-отчёта
       string            m_tff_id_symbol;         // идентификатор инструмента в TFF-отчёте, поле "CFTC Contract Market Code"
       string            m_mt_symbol;
      */
    for(uint idx = 0; idx < this.m_records_size; idx++)
    {
        CTFFRecord curr_record; // выбранная запись
        curr_record.CopyByValue(this.m_records_arr[idx]);
        curr_record.ReverseLongShort();
        this.m_records_arr[idx].CopyByValue(curr_record);
    }
}
//+------------------------------------------------------------------+
//|      Получить индекс в массиве m_records_arr[] по дате           |
//+------------------------------------------------------------------+
int CTFFReport::GetIdxByDate(datetime _date)
{
    int data_idx = WRONG_VALUE;
//---
    if (this.m_records_size > 0)
    {
        data_idx = (int)(this.m_records_size - 1);
        while((data_idx > 0) && (_date >= this.m_records_arr[data_idx - 1].m_tff_date))
            data_idx--;
    }
//---
    return data_idx;
}
//+------------------------------------------------------------------+
//|                    Refresh signal data                           |
//+------------------------------------------------------------------+
bool CTFFReport::RefreshSignalData(const datetime _date, const uint _signal_id, bool _for_open = true)
{
//--- fix some data
    this.m_signal.id = _signal_id;
    if (_for_open)
        this.m_signal.start_date = _date;
    this.m_signal.curr_date = _date;
//---
    SSignalData curr_signal;
    curr_signal.CopyByValue(this.m_signal);
//---
    CTFFRecord curr_record, prev_record;
    int curr_idx = this.GetIdxByDate(curr_signal.curr_date);
//--- проверка по дате сигнала
    /*if (this.m_records_arr[curr_idx].m_tff_date!=_date)
         {
          Print("No report date!");
          return false;
         }
       else*/
    curr_signal.tff_date = this.m_records_arr[curr_idx].m_tff_date;
//---
    curr_signal.idx = curr_idx;

    if (curr_idx < (int)this.m_records_size)
    {
        curr_record.CopyByValue(this.m_records_arr[curr_idx]);
        if (curr_idx < (int)(this.m_records_size - 1))
            prev_record.CopyByValue(this.m_records_arr[curr_idx + 1]);
    }
    else
    {
        Print("Wrong record index!");
        // curr_idx=this.GetIdxByDate(curr_signal.date); // delete
        return false;
    }
// datetime _date=curr_signal.date;
    uint signal_id = curr_signal.id;
    if (curr_signal.id > 16)
        signal_id = curr_signal.id - 16;
//--- grab data
    if ((signal_id <= 2) || (signal_id == 4) || (signal_id == 6) || (signal_id == 7) || (signal_id == 9) || (signal_id == 13) || (signal_id == 16))
    {
        curr_signal.dealer_vals[0] = this.At(_date, TFF_TRADER_DEALER, TFF_TYPE_DATA_LONG_CHG);
        curr_signal.dealer_vals[1] = this.At(_date, TFF_TRADER_DEALER, TFF_TYPE_DATA_SHORT_CHG);
        //---
        curr_signal.asset_vals[0] = this.At(_date, TFF_TRADER_ASSET_MGR, TFF_TYPE_DATA_LONG_CHG);
        curr_signal.asset_vals[1] = this.At(_date, TFF_TRADER_ASSET_MGR, TFF_TYPE_DATA_SHORT_CHG);
        //---
        curr_signal.nonrept_vals[0] = this.At(_date, TFF_TRADER_NONREPT, TFF_TYPE_DATA_LONG_CHG);
        curr_signal.nonrept_vals[1] = this.At(_date, TFF_TRADER_NONREPT, TFF_TYPE_DATA_SHORT_CHG);
        //Print(curr_signal.nonrept_vals[0], " ", curr_signal.nonrept_vals[1]);
    }
    // % of OI
    else if ((!prev_record.m_is_empty) && ((signal_id == 11) || (signal_id == 12) || (signal_id == 14) || (signal_id == 15)))
    {
        double curr_vals[2]; // note: curr_vals[0]-long,curr_vals[1]-short
        double prev_vals[2];
        ArrayInitialize(curr_vals, 0.0);
        ArrayInitialize(prev_vals, 0.0);
        //--- dealer
        for(int i = 0; i < ArraySize(curr_vals); i++)
        {
            /*
                     m_tff_OI_pct_dealer[3];     // dealer position
               double            m_tff_OI_pct_asset_mgr[3];  // asset manager/institutional
               double            m_tff_OI_pct_lev_money[3];  // leveraged funds
               double            m_tff_OI_pct_other_rept[3]; // other reportables
               double            m_tff_OI_pct_nonrept[2];
                    */
            curr_vals[i] = curr_record.m_tff_OI_pct_dealer[i]; // curr_record.m_tff_t_pct_dealer[i];
            prev_vals[i] = prev_record.m_tff_OI_pct_dealer[i]; // prev_record.m_tff_t_pct_dealer[i];
        }
        curr_signal.dealer_vals[0] = curr_vals[0] - prev_vals[0];
        curr_signal.dealer_vals[1] = curr_vals[1] - prev_vals[1];
        //--- asset
        for(int i = 0; i < ArraySize(curr_vals); i++)
        {
            curr_vals[i] = curr_record.m_tff_OI_pct_asset_mgr[i];
            prev_vals[i] = prev_record.m_tff_OI_pct_asset_mgr[i];
        }
        curr_signal.asset_vals[0] = curr_vals[0] - prev_vals[0];
        curr_signal.asset_vals[1] = curr_vals[1] - prev_vals[1];
        //--- leverage
        /*for(int i=0;i<ArraySize(curr_vals);i++)
          {
           curr_vals[i]=curr_record.m_tff_OI_pct_lev_money[i]; // m_tff_t_pct_lev_money[i];
           prev_vals[i]=prev_record.m_tff_OI_pct_lev_money[i];
          }
        curr_signal.lev_vals[0]=curr_vals[0]-prev_vals[0];
        curr_signal.lev_vals[1]=curr_vals[1]-prev_vals[1];*/
    }
    // % of longs
    else if ((!prev_record.m_is_empty) && ((signal_id == 3) || (signal_id == 5) || (signal_id == 8) || (signal_id == 10)))
    {
        double curr_vals[2]; // note: curr_vals[0]-long,curr_vals[1]-short
        double prev_vals[2];
        ArrayInitialize(curr_vals, 0.0);
        ArrayInitialize(prev_vals, 0.0);
        //--- dealer
        for(int i = 0; i < ArraySize(curr_vals); i++)
        {
            curr_vals[i] = curr_record.m_tff_t_pct_dealer[i];
            prev_vals[i] = prev_record.m_tff_t_pct_dealer[i];
        }
        curr_signal.dealer_vals[0] = curr_vals[0] - prev_vals[0];
        curr_signal.dealer_vals[1] = curr_vals[1] - prev_vals[1];
        //--- asset
        for(int i = 0; i < ArraySize(curr_vals); i++)
        {
            curr_vals[i] = curr_record.m_tff_t_pct_asset_mgr[i];
            prev_vals[i] = prev_record.m_tff_t_pct_asset_mgr[i];
        }
        curr_signal.asset_vals[0] = curr_vals[0] - prev_vals[0];
        curr_signal.asset_vals[1] = curr_vals[1] - prev_vals[1];
    }//---
    curr_signal.OI_val = curr_record.m_tff_total_change_OI;
//---
    this.m_signal.CopyByValue(curr_signal);
//---
    return true;
}
//+------------------------------------------------------------------+
//|              Adjust signal initial number                        |
//+------------------------------------------------------------------+
uint CTFFReport::AdjustSignalNum(uint _curr_sig_id)
{
    uint adjusted_id = 0;
//---
// uint curr_signal_id;
    if (_curr_sig_id >= 6 && _curr_sig_id <= 10)
        adjusted_id = _curr_sig_id - 5;
    else if (_curr_sig_id == 16)
        adjusted_id = 13;
    else if (_curr_sig_id > 16)
        adjusted_id = _curr_sig_id - 16;
    else
        adjusted_id = _curr_sig_id;
//--- check
    if ((adjusted_id >= 6 && adjusted_id <= 10) || (adjusted_id >= 16))
        adjusted_id = AdjustSignalNum(adjusted_id);
//---
    return adjusted_id;
}
//+------------------------------------------------------------------+
//|               Получить поле  из строки csv                       |
//+------------------------------------------------------------------+
string CTFFReport::GetFieldCommaDelimited(string _source_string, const uint _field_num)
{
    string ret;
    uint field_cnt = 1;
//---
    int pos1 = 0; // # 1-ой позиции символа ","
    int pos2 = 0; // # 2-ой позиции символа ","
    do
    {
        pos2 = StringFind(_source_string, ",", pos1); //  найти # 1-ой позиции символа ","
        if (pos2 == -1)
            pos2 = StringLen(_source_string);
        ret = StringSubstr(_source_string, pos1, pos2 - pos1); // 1) извлечь подстроку до первого символа ","
        pos1 = pos2 + 1;
        field_cnt++;
    }
    while(field_cnt <= _field_num);
//---
    return ret;
}
//+------------------------------------------------------------------+
//|              Получить запись csv из файла                        |
//+------------------------------------------------------------------+
bool CTFFReport::GetString(int handle, CTFFRecord &_record)
{
    string allstring;
    allstring = FileReadString(handle);
    if (allstring == NULL)
        return true;
//--- имя инструмента в виде "EURO FX - CHICAGO MERCANTILE EXCHANGE"
    string s1 = this.GetFieldCommaDelimited(allstring, 1);

//--- DATA
//--- 1 Market_and_Exchange_Names
    _record.m_tff_symbol = s1;
//--- 3 Report_Date_as_MM_DD_YYYY
    s1 = GetFieldCommaDelimited(allstring, 3); // дата формат 2008-01-08 гггг-мм-дд
    _record.m_tff_date = StringToTime(StringSubstr(s1, 0, 4) + "." + StringSubstr(s1, 5, 2) + "." + StringSubstr(s1, 8, 2));
//--- 4 CFTC_Contract_Market_Code
    s1 = GetFieldCommaDelimited(allstring, 4);
    _record.m_tff_ID = s1;
//--- 8 Open_Interest_All
    s1 = GetFieldCommaDelimited(allstring, 8);
    _record.m_tff_total_OI = StringToDouble(s1);
//--- 9 Dealer_Positions_Long_All,10 Dealer_Positions_Short_All,11 Dealer_Positions_Spread_All
    for(int idx = 0; idx < 3; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 9 + idx);
        _record.m_tff_dealer[idx] = StringToDouble(s1);
    }
//--- 12 Asset_Mgr_Positions_Long_All,13 Asset_Mgr_Positions_Short_All,14 Asset_Mgr_Positions_Spread_All
    for(int idx = 0; idx < 3; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 12 + idx);
        _record.m_tff_asset_mgr[idx] = StringToDouble(s1);
    }
//--- 15 Lev_Money_Positions_Long_All,16 Lev_Money_Positions_Short_All,17 Lev_Money_Positions_Spread_All
    for(int idx = 0; idx < 3; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 15 + idx);
        _record.m_tff_lev_money[idx] = StringToDouble(s1);
    }
//--- 18 Other_Rept_Positions_Long_All,19 Other_Rept_Positions_Short_All,20 Other_Rept_Positions_Spread_All
    for(int idx = 0; idx < 3; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 18 + idx);
        _record.m_tff_other_rept[idx] = StringToDouble(s1);
    }
//--- 23 NonRept_Positions_Long_All,24 NonRept_Positions_Short_All
    for(int idx = 0; idx < 2; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 23 + idx);
        _record.m_tff_nonrept[idx] = StringToDouble(s1);
    }
//--- 25 Change_in_Open_Interest_All
    s1 = GetFieldCommaDelimited(allstring, 25);
    _record.m_tff_total_change_OI = StringToDouble(s1);
//--- 26 Change_in_Dealer_Long_All,27 Change_in_Dealer_Short_All,28 Change_in_Dealer_Spread_All
    for(int idx = 0; idx < 3; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 26 + idx);
        _record.m_tff_change_dealer[idx] = StringToDouble(s1);
    }
//--- 29 Change_in_Asset_Mgr_Long_All,30 Change_in_Asset_Mgr_Short_All,31 Change_in_Asset_Mgr_Spread_All
    for(int idx = 0; idx < 3; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 29 + idx);
        _record.m_tff_change_asset_mgr[idx] = StringToDouble(s1);
    }
//--- 32 Change_in_Lev_Money_Long_All,33 Change_in_Lev_Money_Short_All,34 Change_in_Lev_Money_Spread_All
    for(int idx = 0; idx < 3; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 32 + idx);
        _record.m_tff_change_lev_money[idx] = StringToDouble(s1);
    }
//--- 35 Change_in_Other_Rept_Long_All,36 Change_in_Other_Rept_Short_All,37 Change_in_Other_Rept_Spread_All
    for(int idx = 0; idx < 3; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 35 + idx);
        _record.m_tff_change_other_rept[idx] = StringToDouble(s1);
    }
//--- 40 Change_in_NonRept_Long_All,41 Change_in_NonRept_Short_All
    for(int idx = 0; idx < 2; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 40 + idx);
        _record.m_tff_change_nonrept[idx] = StringToDouble(s1);
    }
//--- 43 Pct_of_OI_Dealer_Long_All,44 Pct_of_OI_Dealer_Short_All,45 Pct_of_OI_Dealer_Spread_All
    for(int idx = 0; idx < 3; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 43 + idx);
        _record.m_tff_OI_pct_dealer[idx] = StringToDouble(s1);
    }
//--- 46 Pct_of_OI_Asset_Mgr_Long_All,47 Pct_of_OI_Asset_Mgr_Short_All,48 Pct_of_OI_Asset_Mgr_Spread_All
    for(int idx = 0; idx < 3; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 46 + idx);
        _record.m_tff_OI_pct_asset_mgr[idx] = StringToDouble(s1);
    }
//--- 49 Pct_of_OI_Lev_Money_Long_All,50 Pct_of_OI_Lev_Money_Short_All,51 Pct_of_OI_Lev_Money_Spread_All
    for(int idx = 0; idx < 3; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 49 + idx);
        _record.m_tff_OI_pct_lev_money[idx] = StringToDouble(s1);
    }
//--- 52 Pct_of_OI_Other_Rept_Long_All,53 Pct_of_OI_Other_Rept_Short_All,54 Pct_of_OI_Other_Rept_Spread_All
    for(int idx = 0; idx < 3; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 52 + idx);
        _record.m_tff_OI_pct_other_rept[idx] = StringToDouble(s1);
    }
//--- 57 Pct_of_OI_NonRept_Long_All,58 Pct_of_OI_NonRept_Short_All
    for(int idx = 0; idx < 2; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 57 + idx);
        _record.m_tff_OI_pct_nonrept[idx] = StringToDouble(s1);
    }
//--- Total Long/Short/spreading contracts (21 Tot_Rept_Positions_Long_All + 23 NonRept_Positions_Long_All)
    for(int idx = 0; idx < 2; idx++)
    {
        s1 = GetFieldCommaDelimited(allstring, 21 + idx);
        double val1 = StringToDouble(s1);
        string s2 = GetFieldCommaDelimited(allstring, 23 + idx);
        double val2 = StringToDouble(s2);
        _record.m_tff_total[idx] = val1 + val2;
    }
//--- Total Spreading contracts
    _record.m_tff_total[2] = _record.m_tff_dealer[2] + _record.m_tff_asset_mgr[2] +
                             _record.m_tff_lev_money[2] + _record.m_tff_other_rept[2];
//---
    for(int idx = 0; idx < 3; idx++)
    {
        double tot = _record.m_tff_total[idx];
        if (tot != 0.0)
        {
            _record.m_tff_t_pct_dealer[idx] = (_record.m_tff_dealer[idx] / tot) * 100.;
            _record.m_tff_t_pct_asset_mgr[idx] = (_record.m_tff_asset_mgr[idx] / tot) * 100.;
            _record.m_tff_t_pct_lev_money[idx] = (_record.m_tff_lev_money[idx] / tot) * 100.;
            _record.m_tff_t_pct_other_rept[idx] = (_record.m_tff_other_rept[idx] / tot) * 100.;
            if (idx < 2)
                _record.m_tff_t_pct_nonrept[idx] = (_record.m_tff_nonrept[idx] / tot) * 100.;
        }
        else
            return false;
    }
    _record.m_is_empty = false;
//---
    return true;
}
//+------------------------------------------------------------------+
//|                Загрузка данных отчета из файла                   |
//+------------------------------------------------------------------+
bool CTFFReport::LoadFile(string _filename)
{
    int handle;
    handle = FileOpen(_filename, FILE_READ | FILE_ANSI | FILE_COMMON);
    if (handle < 1)
    {
        Print("Ошибка открытия файла ", _filename, " ошибка: ", GetLastError());
        return false;
    }
// данные отсортированы в обратном порядке
// пропускаем все до текущего символа
    bool foundsymbol = false;
    bool flagexit = false;
// int cnt=0;
    while(!FileIsEnding(handle) && !flagexit)
    {
        //по строкам файла
        CTFFRecord res_record;
        this.GetString(handle, res_record); // получить строку
        if (StringFind(res_record.m_tff_ID, this.m_tff_id_symbol) != -1)
        {
            //найдена строка с идентификатором инструмента
            this.m_records_size++;
            ArrayResize(this.m_records_arr, this.m_records_size);
            this.m_records_arr[this.m_records_size - 1].CopyByValue(res_record);
            foundsymbol = true;
        }
        else
        {
            if (foundsymbol)
                flagexit = true;
        }
    }
    FileClose(handle);
    return true;
}
//+------------------------------------------------------------------+
//|           Определение идинтификатора в отчете TFF                |
//+------------------------------------------------------------------+
bool CTFFReport::DefineIdSymbol()
{
    this.m_tff_id_symbol = NULL;
//---
    string mt_symbol = this.m_mt_symbol;
    string id = NULL;

//==========================--- Сodes
    if (StringFind(mt_symbol, "USDCAD") >= 0 || StringFind( mt_symbol, "6C") == 0) id = "090741";
    if (StringFind(mt_symbol, "USDCHF") >= 0 || StringFind( mt_symbol, "6S") == 0) id = "092741";
    if (StringFind(mt_symbol, "GBPUSD") >= 0 || StringFind( mt_symbol, "6B") == 0) id = "096742";
    if (StringFind(mt_symbol, "USDJPY") >= 0 || StringFind( mt_symbol, "6J") == 0) id = "097741";
    if (StringFind(mt_symbol, "EURUSD") >= 0 || StringFind( mt_symbol, "6E") == 0) id = "099741";
    if (StringFind(mt_symbol, "EURJPY") >= 0) id = "299661";
    if (StringFind(mt_symbol, "EURGBP") >= 0) id = "599661";
    if (StringFind(mt_symbol, "AUDUSD") >= 0 || StringFind( mt_symbol, "6A") == 0  ) id = "232741";
    if (StringFind(mt_symbol, "USDRUB") >= 0) id = "089741";
    if (StringFind(mt_symbol, "USDMXN") >= 0) id = "095741";
    if (StringFind(mt_symbol, "USDBRL") >= 0) id = "102741";
    if (StringFind(mt_symbol, "NZDUSD") >= 0 || StringFind(mt_symbol, "6N") == 0 ) id = "112741";
//---
    if (mt_symbol == "ZG_CONT" || StringFind(mt_symbol, "ZG") == 0)id = "088606"; //gold
    if (mt_symbol == "ES_CONT" || StringFind( mt_symbol, "ES") == 0 || mt_symbol == "SPX")id = "13874A";
    if (mt_symbol == "GC_CONT" || StringFind(mt_symbol, "GC") == 0 || mt_symbol == "XAUUSD" || mt_symbol == "GOLD")
        id = "088691";
    if (mt_symbol == "XAGUSD" || StringFind(mt_symbol, "SI") == 0)id = "084691"; // серебро comex
//---
//--- todo: ещё?
    if (id == NULL)
    {
        Print("Failed to find the symbol!");
        return false;
    }
    this.m_tff_id_symbol = id;
//---
    return true;
}
//+------------------------------------------------------------------+
//|                         Инициализация                            |
//+------------------------------------------------------------------+
bool CTFFReport::Init(string _term_symbol = NULL)
{
    if (_term_symbol == NULL)
        this.m_mt_symbol = Symbol();
    else
        this.m_mt_symbol = _term_symbol;

//---  определение идентификатора в отчете для текущего символа
    if (!this.DefineIdSymbol())
        return false;
//---
    if ((EndYear - StartYear < 0) || (EndYear - StartYear > 1000))
    {
        Print("Wrong start and end years.");
        return false;
    }
    int year_to_control[];
    int range = EndYear - StartYear + 1;
    ArrayResize(year_to_control, range);
    for (int i = 0; i < range; i++) year_to_control[i] = StartYear + i;

    for (int idx = ArraySize(year_to_control) - 1; idx >= 0; idx--)
    {
        int cury = year_to_control[idx];
        string pref = TFF_STORE_FOLDER + string(cury);
        string filename = pref + "\\FinComYY.txt";
        if (!this.LoadFile(filename))
        {
            Print("Ошибка загрузки файла: " + filename);
            return false;
        }
    }
//--- check base exchange for reversing long and short
    if (this.CheckReverse())
        this.ProcessReverse();
//---
    this.m_signal.Init();
//---
    return true;
}
//+------------------------------------------------------------------+