//+------------------------------------------------------------------+
//|                                         TFF_trader_functions.mqh |
//|                               Copyright 2014-2021, EarnForex.com |
//|                                              Developed by denkir |
//|                           https://login.mql5.com/ru/users/denkir |
//+------------------------------------------------------------------+
#property copyright "EarnForex.com"
#property link      "https://www.earnforex.com/guides/backtesting-strategies-based-on-commitments-of-traders/"
#property version   "1.01"

#include "Fincot.mqh"
#include <User\CisNewBar.mqh>
#include <Trade\Trade.mqh>
//+------------------------------------------------------------------+
//|                        Inputs                                    |
//+------------------------------------------------------------------+
input uint InpSignalNum = 1;  // Trade signal number
input int InpTimeShift1 = -5;  // Time shift (est-gmt)
input int InpTimeShift2 = 3;  // Time shift (MT-gmt)
// input bool InpReverse=false;  // Reverse flag
input double InpLots = 0.1;   // Trade volume
input uint InpTradePause = 350; // Trade pause

//+------------------------------------------------------------------+
//|                        Globals                                   |
//+------------------------------------------------------------------+
CTFFReport myReport;
CTrade myTrade;
CisNewBar myNewBar;
ENUM_ORDER_TYPE open_signal = WRONG_VALUE;
datetime curr_time = 0;
string curr_time_str = NULL;
//====================================================================
//                          Modules
//====================================================================

//+------------------------------------------------------------------+
//|                        Main module                               |
//+------------------------------------------------------------------+
void Main()
{
//--- проверка на возможность торговли и наличия связи
    if (TerminalInfoInteger(TERMINAL_CONNECTED)) // подключение к торговому серверу
        if (TerminalInfoInteger(TERMINAL_TRADE_ALLOWED)) // разрешение на торговлю
            if (MQLInfoInteger(MQL_TRADE_ALLOWED)) // разрешение на торговлю для запущенного советника
                if (!IsStopped())
                {
                    curr_time = TimeCurrent();
                    if (CheckTime(curr_time))
                    {
                        //---
                        curr_time_str = TimeToString(curr_time, TIME_DATE);
                        datetime new_trade_date = StringToTime(curr_time_str);
                        static datetime curr_trade_date;
                        int secs_to_subtract = PeriodSeconds(PERIOD_D1) + (fabs(InpTimeShift1) +
                                               InpTimeShift2) * PeriodSeconds(PERIOD_H1);
                        if ((curr_time - secs_to_subtract) > curr_trade_date)
                        {
                            Print("Новый торговый период: " + TimeToString(new_trade_date, TIME_DATE));
                            curr_trade_date = new_trade_date;
                            //--- типы сигналов
                            ENUM_ORDER_TYPE signals_arr[2];
                            signals_arr[0] = ORDER_TYPE_BUY;
                            signals_arr[1] = ORDER_TYPE_SELL;
                            if (open_signal == ORDER_TYPE_SELL)
                            {
                                signals_arr[0] = ORDER_TYPE_SELL;
                                signals_arr[1] = ORDER_TYPE_BUY;
                            }
                            //---
                            for(int idx = 0; idx < 2; idx++)
                            {
                                ENUM_ORDER_TYPE ord_type = signals_arr[idx];
                                //--- модуль открытия
                                ProcessOpen(ord_type);
                                //--- модуль закрытия
                                ProcessClose(ord_type);
                            }
                        }
                    }
                }
}

//+------------------------------------------------------------------+
//|                       Check time                                 |
//+------------------------------------------------------------------+
bool CheckTime(const datetime _curr_time)
{
    // InpTimeShift1
    bool isTimeToTrade = false;
    datetime gmt_time = _curr_time - InpTimeShift2 * PeriodSeconds(PERIOD_H1);
    datetime est_time = gmt_time + InpTimeShift1 * PeriodSeconds(PERIOD_H1);
//---
    MqlDateTime est_time_s;
    TimeToStruct(est_time, est_time_s);
    if (est_time_s.day_of_week == 5) // Friday
        if (est_time_s.hour >= 15) // hour
            isTimeToTrade = true;
//---
    return isTimeToTrade;
}

//+------------------------------------------------------------------+
//|                   Process open module                            |
//+------------------------------------------------------------------+
void ProcessOpen(ENUM_ORDER_TYPE _init_ord)
{
    if (open_signal != WRONG_VALUE) return;
//---
    ENUM_ORDER_TYPE curr_signal = myReport.CheckOpenSignal(curr_time, InpSignalNum);
    if (InpReverse)
        _init_ord = (_init_ord == ORDER_TYPE_BUY) ? ORDER_TYPE_SELL : ORDER_TYPE_BUY;
    if (curr_signal != WRONG_VALUE)
        if (curr_signal == _init_ord)
        {
            Print("Попытка открыть позицию!");
            //---
            MqlTick last_tick = {0};
            if (SymbolInfoTick(_Symbol, last_tick))
            {
                double curr_pr = (curr_signal == ORDER_TYPE_BUY) ? last_tick.ask : last_tick.bid;
                myTrade.PositionOpen(_Symbol, curr_signal, InpLots, curr_pr, 0.0, 0.0);
                Sleep(InpTradePause);
            }
            //--- есть позиция
            if (PositionSelect(_Symbol))
            {
                open_signal = curr_signal;
                Print("Position open: " + EnumToString(open_signal));
                Print("Position open date: " + TimeToString(curr_time));
            }
        }
}

//+------------------------------------------------------------------+
//|                   Process close module                            |
//+------------------------------------------------------------------+
void ProcessClose(ENUM_ORDER_TYPE _init_ord)
{
    if (open_signal == WRONG_VALUE) return;
//---
    ENUM_ORDER_TYPE curr_signal = myReport.CheckCloseSignal(curr_time);
//---
    if (InpReverse)
        _init_ord = (_init_ord == ORDER_TYPE_BUY) ? ORDER_TYPE_SELL : ORDER_TYPE_BUY;
//---
    if (curr_signal == open_signal)
        if (curr_signal == _init_ord)
        {
            Print("Попытка закрыть позицию!");
            myTrade.PositionClose(_Symbol);
            Sleep(InpTradePause);
            //--- нет позиции
            if (!PositionSelect(_Symbol))
            {
                open_signal = WRONG_VALUE;
                myReport.ResetSignal();
                // Print("Position closed.");
                Print("Position close date: " + TimeToString(curr_time));
            }
        }
}
//+------------------------------------------------------------------+